import json
import os
import pickle
import numpy as np
import pandas as pd

from pathlib import Path
from numpy.random import MT19937, RandomState, SeedSequence


def df_ensemble_stats(dfs):
    """
    Computes the means, medians, and standard deviations column/row-wise over the input
    list of dataframes.

    :param dfs: List of dataframes with identical row/column names.

    :returns: Dictionary of dataframes with the means, medians, and standard deviations.
    """
    df = pd.concat(dfs)
    means = df.groupby(df.index).mean()
    medians = df.groupby(df.index).median()
    stds = df.groupby(df.index).std()

    return {"means": means, "medians": medians, "stds": stds}


def df_stats(df):
    """
    Compute the min, max, mean, median, and standard deviation of the columns in the dataframe.

    :param df: Dataframe.

    :returns: Dataframe of the statistics.
    """
    return pd.DataFrame.from_dict(
        {
            "min": df.min(),
            "max": df.max(),
            "mean": df.mean(),
            "median": df.median(),
            "std": df.std(),
        },
        orient="index",
    )


def filter_df_on_values(df, column_values, drop_filter_columns=True):
    """
    Return a copy of the dataframe filtered conditionally on provided
    column values.

    :param df: Dataframe to filter.
    :param column_values: Dictionary where the keys are column names, and the
        values are values on which to filter the dataframe.
    :param drop_filter_columns: If True returns a copy of the dataframe with
        the filtered columns dropped.

    :returns: A dataframe filtered conditionally on the provided column values.
    """
    df = df.copy()
    for column, value in column_values.items():
        df = df.loc[df[column] == value]

    if drop_filter_columns:
        df.drop(column_values.keys(), axis=1, inplace=True)

    return df


def get_project_dir():
    """
    Gets the project directory path from the environment and checks if it is valid.

    :returns: Path object of the project directory.
    """
    dir_path = os.getenv("QBM_PROJECT_DIR")
    if dir_path is None:
        raise Exception("QBM_PROJECT_DIR env var not set")

    dir_path = Path(dir_path)
    if dir_path.exists():
        return dir_path
    else:
        raise Exception(f"Path '{dir_path}' does not exist")


def get_rng(seed=None):
    """
    Creates a random number generator with the specified seed value.

    :param seed: Seed value for the rng.

    :returns: Numpy RandomState object.
    """
    return RandomState(MT19937(SeedSequence(seed)))


def kl_divergence(p_data, q_data, n_bins=32, epsilon_smooth=None, relative_smooth=False):
    """
    Computes the D_KL(p_data || p_samples).

    Note: this is a crude approximation of the KL divergence.

    :param p_data: Array of data values to compute the p distribution from.
    :param q_data: Array of data values to compute the q distribution from.
    :param n_bins: Number of bins to use in histograms.
    :param epsilon_smooth: Value to use with q distribution smoothing.
    :param relative_smooth: Whether or not the smoothed values are relative to the p
        distribution.

    :returns: D_KL(p || q).
    """
    hist_data, bin_edges = np.histogram(p_data, bins=n_bins)
    hist_samples, _ = np.histogram(q_data, bins=bin_edges)

    p = hist_data / p_data.shape[0]
    q = hist_samples / q_data.shape[0]

    if epsilon_smooth is not None:
        smooth_mask = np.logical_and(p > 0, q == 0)
        not_smooth_mask = np.logical_not(smooth_mask)
        q[smooth_mask] = epsilon_smooth

        if relative_smooth:
            q[smooth_mask] *= p[smooth_mask]

        q[not_smooth_mask] -= q[smooth_mask].sum() / not_smooth_mask.sum()

    assert np.isclose(p.sum(), 1, atol=1e-3)
    assert np.isclose(q.sum(), 1, atol=1e-3)

    support = np.logical_and(p > 0, q > 0)
    p = p[support]
    q = q[support]

    return (p * np.log(p / q)).sum()


def load_artifact(file_path):
    """
    Loads a pickle or json artifact (depending on the file extension).

    :param file_path: Path of the file to load.

    :returns: Loaded python object.
    """
    if type(file_path) == str:
        file_path = Path(file_path)

    if not file_path.exists() or file_path.suffix not in (".json", ".pkl"):
        raise Exception(f"File {file_path} does not exist")

    if file_path.suffix == ".json":
        with open(file_path, "r") as f:
            return json.load(f)
    elif file_path.suffix == ".pkl":
        with open(file_path, "rb") as f:
            return pickle.load(f)


@np.vectorize
def lr_exp_decay(epoch, decay_epoch, period, base=2):
    """
    Exponential decay function for use in learning rate scheduling. It is relative, so one
    must multiply the base learning rate by the output of this function.

    :param epoch: Current epoch.
    :param decay_epoch: Epoch at which to begin the decay.
    :param period: Decay period.
    :param base: Base number of the exponential decay.

    :returns: The learning rate scaling factor.
    """
    return base ** (min((decay_epoch - epoch), 0) / period)


def save_artifact(artifact, file_path):
    """
    Saves a pickle or json artifact (depending on the file extension).

    :param artifact: Python object to save.
    :param file_path: Path of the file to save.
    """
    if type(file_path) == str:
        file_path = Path(file_path)

    if not file_path.parent.exists():
        file_path.parent.mkdir(parents=True)

    if file_path.suffix not in (".json", ".pkl"):
        raise Exception("Invalid file extension")

    if file_path.suffix == ".json":
        with open(file_path, "w") as f:
            json.dump(artifact, f, indent=4)
    elif file_path.suffix == ".pkl":
        with open(file_path, "wb") as f:
            pickle.dump(artifact, f)


def lower_tail_concentration(z, U, V):
    """
    Lower tail concentration function defined as:
    L(z) = P(U <= z | V <= z) = P(U <= z, V <= z) / P(U <= z)
    References:
        - https://freakonometrics.hypotheses.org/2435
        - https://openacttexts.github.io/Loss-Data-Analytics/C-DependenceModel
            (section 14.5.4.3)
        - https://www.casact.org/sites/default/files/old/studynotes_venter_tails_of_copulas.pdf
            (section 3)

    :param z: Tail dependence parameter.
    :param U: Input array for first variable (e.g. X.rank() / (len(X) + 1)).
    :param V: Input array for second variable (e.g. Y.rank() / (len(Y) + 1)).

    :returns: Lower tail concentration function.
    """
    return np.sum(np.logical_and(U <= z, V <= z)) / np.sum(U <= z)


lower_tail_concentration = np.vectorize(lower_tail_concentration, excluded=[1, 2])


def upper_tail_concentration(z, U, V):
    """
    Upper tail concentration function defined as:
    R(z) = P(U > z | V > z) = P(U > z, V > z) / P(U > z)
    References:
        - https://freakonometrics.hypotheses.org/2435
        - https://openacttexts.github.io/Loss-Data-Analytics/C-DependenceModel
            (section 14.5.4.3)
        - https://www.casact.org/sites/default/files/old/studynotes_venter_tails_of_copulas.pdf
            (section 3)

    :param z: Tail dependence parameter.
    :param U: Input array for first variable (e.g. X.rank() / (len(X) + 1)).
    :param V: Input array for second variable (e.g. Y.rank() / (len(Y) + 1)).

    :returns: Upper tail concentration function.
    """
    return np.sum(np.logical_and(U > z, V > z)) / np.sum(U > z)


upper_tail_concentration = np.vectorize(upper_tail_concentration, excluded=[1, 2])



import numpy as np
import pandas as pd


class Discretizer:
    def __init__(self, df, n_bits, epsilon={}):
        """
        :param df: Dataframe of numerical values.
        :param n_bits: Number of bits to discretize to.
        :param epsilon: Optional dictionary of min/max offset values.
        """
        self.columns = df.columns
        self.n_bits = n_bits
        self.epsilon = epsilon
        self.params = {}
        self.split_indices = []
        self.n_bits_total = 0

        for i, column in enumerate(self.columns):
            if column.endswith("_bit"):
                self.params[column] = {"n_bits": 1}
            else:
                self.params[column] = {
                    "n_bits": self.n_bits,
                    "x_min": df[column].min(),
                    "x_max": df[column].max(),
                }
                if column in self.epsilon:
                    self.params[column]["x_min"] -= self.epsilon[column]["min"]
                    self.params[column]["x_max"] += self.epsilon[column]["max"]

            # update the split indices
            if i < len(self.columns) - 1:
                self.split_indices.append(
                    self.params[column]["n_bits"]
                    if i == 0
                    else self.split_indices[i - 1] + self.params[column]["n_bits"]
                )

            self.n_bits_total += self.params[column]["n_bits"]

    @staticmethod
    def bit_vector_to_int(bit_vector):
        """
        Converts a bit vector to a bit string.

        :param bit_vector: Input bit vector.

        :returns: Bit string of the input bit vector.
        """
        return int("".join(str(x) for x in bit_vector), 2)

    @staticmethod
    def bit_vector_to_string(bit_vector):
        """
        Converts a bit vector to a bit string.

        :param bit_vector: Input bit vector.

        :returns: Bit string of the input bit vector.
        """
        return "".join(str(x) for x in bit_vector)

    @staticmethod
    def int_to_bit_vector(x, n_bits):
        """
        Converts the integer x to an n_bits-bit bit vector.

        :param x: Integer value which to convert.
        :param n_bits: Length of the bit vector.

        :returns: Bit vector of length n_bits.
        """
        return [1 if i == "1" else 0 for i in bin(x)[2:].zfill(n_bits)]

    @staticmethod
    @np.vectorize
    def discretize(x, n_bits, x_min, x_max):
        """
        Convert the value x into a n-bit bit string.

        :param x: Float value to convert.
        :param n_bits: Length of the bit string.
        :param x_min: Minimum value for scaling.
        :param x_max: Maximum value for scaling.

        :returns: A bit string representation of x.
        """
        scaling_factor = (2 ** n_bits - 1) / (x_max - x_min)

        x = round((x - x_min) * scaling_factor)
        assert x >= 0 and x <= 2 ** n_bits - 1
        return x

    @staticmethod
    @np.vectorize
    def undiscretize(x, n_bits, x_min, x_max):
        """
        Convert the value x into a float from a n-bit bit string.

        :param x: Int value to convert.
        :param n_bits: Length of the bit string.
        :param x_min: Minimum value for scaling.
        :param x_max: Maximum value for scaling.

        :returns: A float representation of x.
        """
        scaling_factor = (2 ** n_bits - 1) / (x_max - x_min)

        assert x < 2 ** n_bits
        return x / scaling_factor + x_min

    def discretize_df(self, df):
        """
        Convert all columns of a dataframe to bit representation.

        :param df: Dataframe which to convert.

        :returns: A discretized version of df.
        """
        df_discretized = df.copy()
        for column in df.columns:
            if column.endswith("_bit"):
                df_discretized[column] = df[column].astype(np.int8)
            else:
                df_discretized[column] = self.discretize(df[column], **self.params[column])

        return df_discretized

    def undiscretize_df(self, df):
        """
        Convert all columns of a dataframe to floats from bit representation.

        :param df: Dataframe which to convert.

        :returns: An undiscretized version of df_discretized.
        """
        df_undiscretized = df.copy()
        for column in df.columns:
            if column.endswith("_bit"):
                df_undiscretized[column] = df[column].astype(np.int8)
            else:
                df_undiscretized[column] = self.undiscretize(
                    df[column], **self.params[column]
                )

        return df_undiscretized

    def df_to_bit_array(self, df):
        """
        Converts a dataframe of floats to a bit array.

        :param df: Dataframe which to convert.

        :param returns: Array of bits of shape (df.shape[0], self.n_bits_total).
        """
        assert set(self.columns) == set(df.columns)

        df = self.discretize_df(df)
        bit_array = np.hstack(
            [
                np.vstack(
                    [
                        self.int_to_bit_vector(x, self.params[column]["n_bits"])
                        for x in df[column]
                    ]
                )
                for column in self.columns
            ]
        )

        return bit_array

    def bit_array_to_df(self, bit_array):
        """
        Converts bit array a dataframe of floats.

        :param bit_array: Bit array which to convert.

        :param returns: Dataframe of shape (bit_array.shape[0], len(self.columns)).
        """
        assert len(bit_array[0]) == self.n_bits_total

        rows = [
            [
                self.bit_vector_to_int(x)
                for x in np.array_split(bit_vector, self.split_indices)
            ]
            for bit_vector in bit_array
        ]
        df = self.undiscretize_df(pd.DataFrame(rows, columns=self.columns))

        return df



import numpy as np
from scipy.sparse import csr_matrix, kron, identity, diags
from scipy.linalg import eigh


# set constants
sparse_X = csr_matrix(([1, 1], ([0, 1], [1, 0])), dtype=np.float64)
sparse_Z = csr_matrix(([1, -1], ([0, 1], [0, 1])), dtype=np.float64)


def get_pauli_kron(n_visible, n_hidden):
    """
    Computes the necessary Pauli Kronecker product (sparse) matrices for a n_visible +
    n_hidden qubit problem. Used as an argument to compute_H, e.g. one would instantiate
    pauli_kron as pauli_kron = get_pauli_kron(n_visible, n_hidden), then pass to compute_H
    when computing the Hamiltonian.

    :param n_visible: Number of visible units.
    :param n_hidden: Number of hidden units.

    :returns: A dictionary of Kronecker product Pauli matrices.
    """
    # set Kronecker product Pauli matrices
    n_qubits = n_visible + n_hidden
    pauli_kron = {}
    for i in range(n_qubits):
        pauli_kron["x", i] = sparse_kron(i, n_qubits, sparse_X)
        pauli_kron["z_diag", i] = sparse_kron(i, n_qubits, sparse_Z).diagonal()
    for i in range(n_qubits):
        for j in range(i + 1, n_qubits):
            pauli_kron["zz_diag", i, j] = pauli_kron["z_diag", i] * pauli_kron["z_diag", j]

    return pauli_kron


def sparse_kron(i, n_qubits, A):
    """
    Compute I_{2^i} ⊗ A ⊗ I_{2^(n_qubits-i-1)}.

    :param i: Index of the "A" matrix.
    :param n_qubits: Total number of qubits.
    :param A: Matrix to tensor with identities.

    :returns: I_{2^i} ⊗ A ⊗ I_{2^(n_qubits-i-1)}.
    """
    if i != 0 and i != n_qubits - 1:
        return kron(kron(identity(2 ** i), A), identity(2 ** (n_qubits - i - 1)))
    if i == 0:
        return kron(A, identity(2 ** (n_qubits - 1)))
    if i == n_qubits - 1:
        return kron(identity(2 ** (n_qubits - 1)), A)


def compute_H(h, J, A, B, n_qubits, pauli_kron):
    """
    Computes the Hamiltonian of the annealer at relative time s.

    :param h: Linear Ising terms.
    :param J: Quadratic Ising terms.
    :param A: Coefficient of the off-diagonal terms, e.g. A(s).
    :param B: Coefficient of the diagonal terms, e.g. B(s).
    :param n_qubits: Number of qubits.
    :param pauli_kron: Kronecker product Pauli matrices dict.

    :returns: Hamiltonian matrix H.
    """
    # diagonal terms
    H_diag = np.zeros(2 ** n_qubits)
    for i in range(n_qubits):
        # linear terms
        if h[i] != 0:
            H_diag += (B * h[i]) * pauli_kron["z_diag", i]

        # quadratic terms
        for j in range(i + 1, n_qubits):
            if J[i, j] != 0:
                H_diag += (B * J[i, j]) * pauli_kron["zz_diag", i, j]

    # return just the diagonal if H is a diagonal matrix
    if A == 0:
        return np.diag(H_diag)

    # off-diagonal terms
    H = csr_matrix((2 ** n_qubits, 2 ** n_qubits), dtype=np.float64)
    for i in range(n_qubits):
        H -= A * pauli_kron["x", i]

    return (H + diags(H_diag, format="csr")).toarray()


def compute_rho(H, beta, diagonal=False):
    """
    Computes the trace normalized density matrix rho.

    :param H: Hamiltonian matrix.
    :param beta: Inverse temperature beta = 1 / (k_B * T).
    :param diagonal: Flag to indicate whether H is a diagonal matrix or not.

    :return: Density matrix rho.
    """
    # if diagonal then compute directly, else use eigen decomposition
    if diagonal:
        Lambda = H.diagonal()
        exp_beta_Lambda = np.exp(-beta * (Lambda - Lambda.min()))
        return np.diag(exp_beta_Lambda / exp_beta_Lambda.sum())
    else:
        Lambda, S = eigh(H)
        exp_beta_Lambda = np.exp(-beta * (Lambda - Lambda.min()))
        return (S * (exp_beta_Lambda / exp_beta_Lambda.sum())) @ S.T



from .discretization import Discretizer
from .misc import (
    df_stats,
    df_ensemble_stats,
    filter_df_on_values,
    get_project_dir,
    get_rng,
    kl_divergence,
    load_artifact,
    lr_exp_decay,
    save_artifact,
    lower_tail_concentration,
    upper_tail_concentration,
)
from .transformations import PowerTransformer

__all__ = [
    # discretization
    "Discretizer",
    # misc.
    "df_stats",
    "df_ensemble_stats",
    "filter_df_on_values",
    "get_project_dir",
    "get_rng",
    "kl_divergence",
    "load_artifact",
    "lr_exp_decay",
    "save_artifact",
    "lower_tail_concentration",
    "upper_tail_concentration",
    # transformations
    "PowerTransformer",
]



from abc import ABC, abstractmethod

import numpy as np

from qbm.utils import get_rng


class QBMBase(ABC):
    """
    Abstract base class for Quantum Boltzmann Machines

    Theory based on Quantum Boltzmann Machine by Amin et al.
    https://journals.aps.org/prx/abstract/10.1103/PhysRevX.8.021050
    """

    def __init__(self, V_train, n_hidden, seed):
        """
        :param V_train: Training data.
        :param n_hidden: Number of hidden units.
        :param seed: Seed for the random number generator.
        """
        self.V_train = V_train
        self.n_visible = V_train.shape[1]
        self.n_hidden = n_hidden
        self.n_qubits = self.n_visible + self.n_hidden
        self.seed = seed
        self.rng = get_rng(self.seed)
        self.grads = {}

        self._initialize_weights_and_biases()

    def _apply_grads(self, learning_rate):
        """
        Applies the gradients from the positive and negative phases using the provided
        learning rate.

        :param learning_rate: Learning rate to scale the gradients with.
        """
        self.b += learning_rate * (self.grads["b_pos"] - self.grads["b_neg"])
        self.W += learning_rate * (self.grads["W_pos"] - self.grads["W_neg"])

    def _binary_to_eigen(self, x):
        """
        Convert bit values {0, 1} to corresponding spin values {+1, -1}.

        :param x: Input array of values {0, 1}.

        :returns: Output array of values {+1, -1}.
        """
        return (1 - 2 * x).astype(np.int8)

    def _eigen_to_binary(self, x):
        """
        Convert spin values {+1, -1} to corresponding bit values {0, 1}.

        :param x: Input array of values {+1, -1}.

        :returns: Output array of values {0, 1}.
        """
        return ((1 - x) / 2).astype(np.int8)

    def _random_mini_batch_indices(self, mini_batch_size):
        """
        Generates random, non-intersecting sets of indices for creating mini-batches of the
        training data.

        :param mini_batch_size: Size of the mini-batches.

        :returns: List of numpy arrays, each array containing the indices corresponding to
            a mini-batch.
        """
        return np.split(
            self.rng.permutation(np.arange(self.V_train.shape[0])),
            np.arange(mini_batch_size, self.V_train.shape[0], mini_batch_size),
        )

    @abstractmethod
    def _compute_positive_grads(self):
        pass

    @abstractmethod
    def _compute_negative_grads(self):
        pass

    @abstractmethod
    def _initialize_weights_and_biases(self):
        pass



import numpy as np


class PowerTransformer:
    """
    Transforms data points that lie beyond the provided threshold by a taking their
    power (<1) to scale them closer to the mean.
    """

    def __init__(self, df, threshold=1, power=0.5, columns=None):
        """
        :param df: Dataframe of data to scale.
        :param threshold: Number of standard deviations above the mean at which to begin
            the scaling.
        :param power: Power at which to scale the outlier.
        :param columns: Optional list of columns to apply the transformation to. If no
            columns are provided, then all columns are transformed.
        """
        assert power < 1
        assert threshold >= 1

        if columns is None:
            self.columns = df.columns
        else:
            self.columns = columns
        self.power = power
        self.threshold = threshold
        self.offset = threshold - threshold ** power

        self.μ = {}
        self.σ = {}
        for column in df.columns:
            self.μ[column] = df[column].mean()
            self.σ[column] = df[column].std()

    def transform(self, df, inplace=False):
        """
        Transforms the data to the scaled space.

        :param df: Dataframe to scale.
        :param inplace: If True then it operates on the same dataframe, if False then
            it creates a copy.

        :returns: Dataframe of transformed data (if inplace == False).
        """
        if not inplace:
            df = df.copy()

        for column in self.columns:
            μ = self.μ[column]
            σ = self.σ[column]
            x = (df[column] - μ) / σ
            mask = np.abs(x) > self.threshold
            x[mask] = ((np.abs(x) ** self.power + self.offset) * np.sign(x))[mask]
            df[column] = x * σ + μ

        if not inplace:
            return df

    def inverse_transform(self, df, inplace=False):
        """
        Transforms the data back from the scaled space.

        :param df: Dataframe to scale.
        :param inplace: If True then it operates on the same dataframe, if False then
            it creates a copy.

        :returns: Dataframe of untransformed data (if inplace == False).
        """
        if not inplace:
            df = df.copy()

        for column in self.columns:
            μ = self.μ[column]
            σ = self.σ[column]
            x = (df[column] - μ) / σ
            mask = np.abs(x) > self.threshold
            x[mask] = ((np.abs(x) - self.offset) ** (1 / self.power) * np.sign(x))[mask]
            df[column] = x * σ + μ

        if not inplace:
            return df



from datetime import timedelta
from time import time

import numpy as np
from dwave.system import DWaveSampler, FixedEmbeddingComposite

from qbm.models import QBMBase
from qbm.utils import Discretizer, load_artifact, save_artifact
from qbm.simulation import compute_H, compute_rho, get_pauli_kron


class BQRBM(QBMBase):
    """
    Bound-based Quantum Restricted Boltzmann Machine
    """

    def __init__(
        self,
        V_train,
        n_hidden,
        A_freeze,
        B_freeze,
        beta_initial=1.0,
        beta_range=[0.1, 10],
        annealer_params=None,
        simulation_params=None,
        seed=0,
    ):
        """
        Note: one of either annealer_params or simulation_params must not be None. Whichever
        is not None determines whether the samples are generated by the annealer or the
        simulation, respectively.

        The simulation works by exact computation of ρ(s, T) = e^{-β * H(s)} / Z. H(s) is
        determined by A_freeze and B_freeze.

        :param V_train: Training data set of visible vectors, shape (n_samples, n_visible).
        :param n_hidden: Number of hidden units.
        :param A_freeze: Value of A(s) at the freeze out point. Used for Γ = β * A_freeze.
            Units in GHz.
        :param B_freeze: Value of B(s) at the freeze out point. Used for
            b_i = -β * B_freeze * h_i and w_ij = -β * B_freeze * J_ij. Units in GHz.
        :param beta_initial: Initial value of the effective β. Units in 1/GHz.
        :param beta_range: Range of allowed β values, used for making sure β is not updated
            to an infeasible value (e.g. negative).
        :param annealer_params: Dictionary with keys:
            - "schedule": List of (t, s) tuples defining the anneal schedule.
            - "embedding": Dict mapping the logical to physical qubits.
            - "relative_chain_strength" [optional]: Relative chain strength value.
            - "qpu_params" [optional]: Parameters dict to unpack to DWaveSampler(), e.g.
                {"region": "na-west-1", "solver": "Advantage_system4.1"}
        :param simulation_params: Dictionary with keys:
            - "beta": Effective β that the simulation generates samples at.
        :param seed: Seed for the random number generator. Used for random minibatches, as
            well as the exact sampler.
        """
        # convert from binary to ±1 if necessary
        if set(np.unique(V_train)) == set([0, 1]):
            V_train = self._binary_to_eigen(V_train)
        assert set(np.unique(V_train)) == set([-1, 1])

        self.A_freeze = A_freeze
        self.B_freeze = B_freeze
        self.beta = beta_initial
        self.beta_range = beta_range
        self.beta_history = [beta_initial]
        super().__init__(V_train=V_train, n_hidden=n_hidden, seed=seed)

        # check if requirements met to use annealer or simulation
        if (annealer_params is not None and simulation_params is not None) or (
            annealer_params is None and simulation_params is None
        ):
            raise Exception(
                "You must pass one of either annealer_params or simulation_params, not both"
            )

        elif annealer_params is not None:
            annealer_params_keys = ["schedule", "embedding"]
            for k in annealer_params_keys:
                if k not in annealer_params:
                    raise Exception(
                        f"Missing key in annealer_params. Required keys are {annealer_params_keys}"
                    )

            self.annealer_params = annealer_params
            self._initialize_annealer()

        elif simulation_params is not None:
            simulation_params_keys = ["beta"]
            for k in simulation_params_keys:
                if k not in simulation_params:
                    raise Exception(
                        f"Missing key in simulation_params. Required keys are {simulation_params_keys}"
                    )

            self.simulation_params = simulation_params
            self._pauli_kron = get_pauli_kron(self.n_visible, self.n_hidden)
            self.h_range = np.array(
                self.simulation_params.get("h_range", [-np.inf, np.inf])
            )
            self.J_range = np.array(
                self.simulation_params.get("J_range", [-np.inf, np.inf])
            )

    def sample(self, n_samples, answer_mode="raw", use_gauge=True, binary=False):
        """
        Generate samples using the model, either exact or from the annealer.

        :param n_samples: Number of samples to generate (num_reads param in sample_ising).
        :param answer_mode: "raw" or "histogram".
        :param use_gauge: If True will use a random gauge transformation (recommended for
            more robust sample generation).
        :param binary: If true will convert the state vector values from {+1, -1} to {0, 1}.

        :returns: Dictionary (exact) or Ocean SDK SampleSet object (annealer).
        """
        if hasattr(self, "simulation_params"):
            return self._sample_simulation(n_samples, binary=binary)
        else:
            return self._sample_annealer(
                n_samples, answer_mode=answer_mode, use_gauge=use_gauge, binary=binary
            )

    def train(
        self,
        n_epochs=100,
        learning_rate=1e-1,
        learning_rate_beta=1e-1,
        mini_batch_size=10,
        n_samples=10_000,
        callback=None,
    ):
        """
        Fits the model to the training data.

        :param n_epochs: Number of epochs to train for.
        :param learning_rate: Learning rate. If a list or array, then it will represent the
            learning rate over the epochs, must be of length n_epochs.
        :param learning_rate_beta: Learning rate for the effective temperature. If a list or
            array, then it will represent the learning rate over the epochs, must be of
            length n_epochs.
            Note: It might be useful to use a larger learning_rate_beta in the beginning to
            help the model find a good temperature, then drop it after a number of epochs.
        :param mini_batch_size: Size of the mini-batches.
        :param n_samples: Number of samples to generate after every epoch. Used for
            computing β gradient, as well as the callback.
        :param callback: A function called at the end of each epoch. It takes the arguments
            (model, samples), and returns a dictionary with required keys ["value",
            "print"], where the "print" value is a string to be printed at the end of each
            epoch.
        """
        if isinstance(learning_rate, float):
            learning_rate = [learning_rate] * n_epochs
        assert len(learning_rate) == n_epochs

        if isinstance(learning_rate_beta, float):
            learning_rate_beta = [learning_rate_beta] * n_epochs
        assert len(learning_rate_beta) == n_epochs

        if not hasattr(self, "callback_history"):
            self.callback_history = []

        for epoch in range(1, n_epochs + 1):
            start_time = time()

            # set the effective learning rates
            self.learning_rate = learning_rate[epoch - 1]
            self.learning_rate_beta = learning_rate_beta[epoch - 1]

            # compute and apply gradient updates for each mini batch
            for mini_batch_indices in self._random_mini_batch_indices(mini_batch_size):
                V_pos = self.V_train[mini_batch_indices]
                self._compute_positive_grads(V_pos)
                self._compute_negative_grads(V_pos.shape[0])
                self._apply_grads(self.learning_rate / V_pos.shape[0])
                self._check_h_and_H_ranges()

            # update β
            samples = self.sample(n_samples)
            self._update_beta(samples)
            self._check_h_and_H_ranges()

            # callback function
            if callback is not None:
                callback_output = callback(self, self._get_state_vectors(samples))
                self.callback_history.append(callback_output)

            # print diagnostics
            end_time = time()
            print(
                f"[{type(self).__name__}] epoch {epoch}:",
                f"β = {self.beta:.3f},",
                f"learning rate = {learning_rate[epoch - 1]:.2e},",
                f"β learning rate = {learning_rate_beta[epoch - 1]:.2e},",
                f"epoch duration = {timedelta(seconds=end_time - start_time)}",
            )
            if callback is not None and "print" in callback_output:
                print(callback_output["print"])

    def save(self, file_path, reinitialize_annealer=True):
        """
        Saves the BQRBM model at file_path. Necessary because of pickling issues with the
        qpu and sampler objects.

        :param file_path: Path to save the model to. Must be a Path object or a string with
            ".pkl" file extension.
        :param reinitialize_annealer: If True and has attribute self.annealer_params, then
            will call self._initialize_annealer() after saving.
        """
        if hasattr(self, "annealer_params"):
            self.qpu = None
            self.sampler = None

        save_artifact(self, file_path)

        if hasattr(self, "annealer_params") and reinitialize_annealer:
            self._initialize_annealer()

    @staticmethod
    def load(file_path, initialize_annealer=True):
        """
        Loads the BQRBM model at file_path. Necessary because of pickling issues with the
        qpu and sampler objects.

        :param file_path: Path to the model to load. Must be a Path object or a string with
            ".pkl" file extension.
        :param initialize_annealer: If True and has attribute self.annealer_params, then
            will call self._initialize_annealer().

        :returns: BQRBM instance loaded from the file path.
        """
        model = load_artifact(file_path)
        if hasattr(model, "annealer_params") and initialize_annealer:
            model._initialize_annealer()

        return model

    @property
    def h(self):
        """
        Ising h values. Correspond to b_i = -β * B_freeze * h_i
        """
        return -self.b / (self.beta * self.B_freeze)

    @property
    def J(self):
        """
        Ising J values. Correspond to w_ij = -β * B_freeze * J_ij
        """
        J = np.zeros((self.n_qubits, self.n_qubits))
        J[: self.n_visible, self.n_visible :] = -self.W / (self.beta * self.B_freeze)
        return J

    def _check_h_and_H_ranges(self):
        """
        Raises and exception if h and J values do not fall within h_range and J_range.
        """
        h_satisfied = np.logical_and(
            self.h > self.h_range.min(), self.h < self.h_range.max()
        ).all()
        J_satisfied = np.logical_and(
            self.J > self.J_range.min(), self.J < self.J_range.max()
        ).all()

        if not h_satisfied or not J_satisfied:
            raise Exception("Learned h and J values outside of allowed range")

    def _compute_positive_grads(self, V_pos):
        """
        Computes the gradients for the positive phase, i.e., the expectation values w.r.t.
        the clamped Hamiltonian.

        :param V_pos: Training data set mini-batch of positive vectors, shape
            (mini_batch_size, n_visible).
        """
        b_hidden = self.b[self.n_visible :] + V_pos @ self.W
        D = np.sqrt((self.beta * self.A_freeze) ** 2 + b_hidden ** 2)
        H_pos = (b_hidden / D) * np.tanh(D)

        self.grads["b_pos"] = np.concatenate((V_pos.mean(axis=0), H_pos.mean(axis=0)))
        self.grads["W_pos"] = V_pos.T @ H_pos / V_pos.shape[0]

    def _compute_negative_grads(self, n_samples):
        """
        Computes the gradients for the negative phase, i.e., the expectation values w.r.t.
        the model distribution.

        :param n_samples: Number of samples to use in the negative phase.
        """
        samples = self.sample(n_samples)
        state_vectors = self._get_state_vectors(samples)

        V_neg = state_vectors[:, : self.n_visible]
        b_hidden = self.b[self.n_visible :] + V_neg @ self.W
        D = np.sqrt((self.beta * self.A_freeze) ** 2 + b_hidden ** 2)
        H_neg = (b_hidden / D) * np.tanh(D)

        self.grads["b_neg"] = np.concatenate((V_neg.mean(axis=0), H_neg.mean(axis=0)))
        self.grads["W_neg"] = V_neg.T @ H_neg / V_neg.shape[0]

    def _get_state_vectors(self, samples):
        """
        Get the state vectors from the samples (depending on exact or annealer generated).

        :param samples: Return value out of BQRBM.sample().

        :returns: Array of state vectors, shape (n_samples, n_qubits).
        """
        if hasattr(self, "simulation_params"):
            return samples["state_vectors"]
        else:
            return samples.record.sample

    def _initialize_annealer(self):
        """
        Initializes the D-Wave sampler using the fixed embedding provided to the object
        instantiation.
        """
        self.qpu = DWaveSampler(**self.annealer_params.get("qpu_params", {}))
        self.sampler = FixedEmbeddingComposite(self.qpu, self.annealer_params["embedding"])
        self.h_range = np.array(self.qpu.properties["h_range"])
        self.J_range = np.array(self.qpu.properties["j_range"])

    def _initialize_weights_and_biases(self, mu=0, sigma=0.1):
        """
        Initializes the weights and biases.

        :param mu: Mean of the normal distribution of the weights.
        :param sigma: Standard deviation of the normal distribution of the weights.
        """
        self.b = np.zeros(self.n_qubits)
        self.W = self.rng.normal(mu, sigma, (self.n_visible, self.n_hidden))

    def _mean_classical_energy(self, V, H, VW):
        """
        Computes the mean classical energy w.r.t. the weights and biases over the provided
        visible and hidden unit state vectors.

        :param V: Numpy array where the rows are visible units.
        :param H: Numpy array where the rows are hidden units.
        :param VW: V @ W (used to avoid double computation).

        :returns: Mean energy.
        """
        return (
            -(V @ self.b[: self.n_visible]).sum()
            - (H @ self.b[self.n_visible :]).sum()
            - np.einsum("kj,kj", VW, H)
        ) / V.shape[0]

    def _sample_annealer(self, n_samples, answer_mode="raw", use_gauge=True, binary=False):
        """
        Obtain a sample set using the annealer.

        :param n_samples: Number of samples to generate (num_reads param in sample_ising).
        :param answer_mode: "raw" or "histogram".
        :param use_gauge: If True will use a random gauge transformation (recommended for
            more robust sample generation).
        :param binary: If true will convert the state vector values from {-1, +1} to {0, 1}.

        :returns: Ocean SDK SampleSet object.
        """
        # compute the h's and J's
        h = self.h
        J = self.J

        # apply a random gauge
        if use_gauge:
            gauge = self.rng.choice([-1, 1], self.n_qubits)
            h *= gauge
            J *= np.outer(gauge, gauge)

        # compute the chain strength
        chain_strength = self.annealer_params.get("relative_chain_strength")
        if chain_strength is not None:
            chain_strength *= max(np.abs(h).max(), np.abs(J).max())
            chain_strength = min(chain_strength, self.J_range.max())

        # get samples from the annealer
        samples = self.sampler.sample_ising(
            h,
            J,
            num_reads=n_samples,
            anneal_schedule=self.annealer_params.get("schedule"),
            chain_strength=chain_strength,
            answer_mode=answer_mode,
            auto_scale=False,
        )

        # undo the gauge
        if use_gauge:
            samples.record.sample *= gauge

        # convert to binary if specified
        if binary:
            samples.record.sample = self._eigen_to_binary(samples.record.sample)

        return samples

    def _sample_simulation(self, n_samples, binary=False):
        """
        Sample using the exact computed probabilities.

        :param n_samples: Number of samples to generate.
        :param binary: If true will convert the state vector values from {-1, +1} to {0, 1}.

        :returns: Dict with keys:
            - "E": Energies, i.e., the diagonal of the Hamiltonian.
            - "p": Probabilities, i.e., the diagonal of the density matrix.
            - "states": Integer state numbers of sampled states.
            - "state_vectors": Array of sampled state vectors, shape (n_samples, n_qubits).
        """
        # compute the h's and J's
        h = self.h
        J = self.J

        # compute the Hamiltonian and density matrix
        H = compute_H(h, J, self.A_freeze, self.B_freeze, self.n_qubits, self._pauli_kron)
        rho = compute_rho(H, self.simulation_params["beta"], diagonal=(self.A_freeze == 0))

        # sample using the probabilities on the diagonal of rho
        samples = {}
        samples["E"] = np.diag(H).copy()
        samples["p"] = np.diag(rho).copy()
        samples["states"] = self.rng.choice(
            range(2 ** self.n_qubits), size=n_samples, p=samples["p"]
        )
        samples["state_vectors"] = self._binary_to_eigen(
            np.vstack(
                [Discretizer.int_to_bit_vector(x, self.n_qubits) for x in samples["states"]]
            )
        )

        # convert to binary if specified
        if binary:
            samples["state_vectors"] = self._eigen_to_binary(samples["state_vectors"])

        return samples

    def _update_beta(self, samples):
        """
        Updates the effective β = 1 / kT estimator. Used for scaling the coefficients sent
        to the annealer.

        Note: in its current form, this method only works when s_freeze = 1, i.e., when
        training and sampling from classical Boltzmann distributions.

        :param samples: Samples to use for computing the mean energies w.r.t. the model.
        """
        # compute the train energy
        VW_train = self.V_train @ self.W
        b_eff = self.b[self.n_visible :] + VW_train
        D = np.sqrt((self.beta * self.A_freeze) ** 2 + b_eff ** 2)
        H_train = (b_eff / D) * np.tanh(D)
        E_train = self._mean_classical_energy(self.V_train, H_train, VW_train)

        # compute the model energy
        state_vectors = self._get_state_vectors(samples)
        V_model = state_vectors[:, : self.n_visible]
        VW_model = V_model @ self.W
        H_model = state_vectors[:, self.n_visible :]
        E_model = self._mean_classical_energy(V_model, H_model, VW_model)

        # update the params
        self.beta = np.clip(
            self.beta + self.learning_rate_beta * (E_train - E_model), *self.beta_range
        )
        self.beta_history.append(self.beta)



import setuptools

# with open("requirements.txt", "r") as requirements_file:
# requirements = requirements_file.read().splitlines()

requirements = [
    "dwave-ocean-sdk>=4.2.0",
    "matplotlib>=3.4.3",
    "numpy>=1.20.3",
    "pandas>=1.3.3",
    "scipy>=1.7.1",
]


setuptools.setup(install_requires=requirements)
